@@ -41,9 +41,15 @@ typedef struct _RdpPeerContext
   /* Virtual Channel Manager */
   HANDLE vcm;
 
+  GrdRdpDvc *rdp_dvc;
+
+  GMutex channel_mutex;
+
   GrdClipboardRdp *clipboard_rdp;
+  GrdRdpAudioPlayback *audio_playback;
   GrdRdpDisplayControl *display_control;
   GrdRdpGraphicsPipeline *graphics_pipeline;
+  GrdRdpTelemetry *telemetry;
 } RdpPeerContext;
 
 #endif /* GRD_RDP_PRIVATE_H */
